﻿/*
	JTVPing - twitch.tv ingest server latency monitor
    Copyright (C) 2012 Richard Stanway, teamliquid.net

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;
using System.Xml;
using System.Collections;

namespace JTVPing
{
	public partial class FormMain : Form
    {
		public string[] serverAddresses;
        public TextBox[]   textHosts;
        public Label[]     labelAvg;
        public Label[]     labelLast;
        public Label[]     labelJitter;
		public Sample		sampler;
		myTwitchServer[] twitchServers;

		private int NUM_SERVERS = 0;

        public FormMain()
        {
			InitializeComponent();

			backgroundWorker2.RunWorkerAsync();
        }

		public partial class Sample
		{
			private int[][] mSecs;
			private int[] nSamples;

			private const int MAX_SAMPLES = 300;

			public Sample(int num_servers)
			{
				mSecs = new int[num_servers][];
				nSamples = new int[num_servers];

				for (int i = 0; i < mSecs.Length; i++)
					mSecs[i] = new int[MAX_SAMPLES];
			}

			private static double StandardDeviation(int[] valueList, int numElements)
			{
				double M = 0.0;
				double S = 0.0;
				int k = 1;

				for (int i = 0; i < numElements; i++)
				{
					double value = valueList[i];
					double tmpM = M;
					M += (value - tmpM) / k;
					S += (value - tmpM) * (value - M);
					k++;
				}
				return Math.Sqrt(S / (k - 1));
			}

			public void UpdateStatus (FormMain f)
			{
				for (int i = 0; i < mSecs.Length; i++)
				{
					if (nSamples[i] > 0)
					{
						int	j;
						int total;

						if (mSecs[i][(nSamples[i] - 1) % MAX_SAMPLES] < 75)
							f.labelLast[i].ForeColor = Color.Green;
						else if (mSecs[i][(nSamples[i] - 1) % MAX_SAMPLES] < 150)
							f.labelLast[i].ForeColor = Color.Olive;
						else
							f.labelLast[i].ForeColor = Color.Red;
						f.labelLast[i].Text = mSecs[i][(nSamples[i] - 1) % MAX_SAMPLES] + "ms";

						total = 0;
						for (j = 0; j < nSamples[i]; j++)
						{
							if (j == MAX_SAMPLES)
								break;
							total += mSecs[i][j];
						}

						int avg = (int)(total / j);
						if (avg < 75)
							f.labelAvg[i].ForeColor = Color.Green;
						else if (avg < 150)
							f.labelAvg[i].ForeColor = Color.Olive;
						else
							f.labelAvg[i].ForeColor = Color.Red;

						f.labelAvg[i].Text = avg + "ms";

						double sdev = Math.Round(StandardDeviation(mSecs[i], j), 2);
						if (sdev < 7.5)
							f.labelJitter[i].ForeColor = Color.Green;
						else if (sdev < 20)
							f.labelJitter[i].ForeColor = Color.Olive;
						else
							f.labelJitter[i].ForeColor = Color.Red;
							
						f.labelJitter[i].Text = sdev + "ms";
					}
				}
			}

			public void AddSample(int index, int msecs)
			{
				mSecs[index][nSamples[index] % MAX_SAMPLES] = msecs;
				nSamples[index]++;
			}
		}

		public static void ConnectCallback(IAsyncResult ar)
		{
			if (ar.IsCompleted)
			{
				ManualResetEvent ev = (ManualResetEvent)ar.AsyncState;
				ev.Set();
			}
		}

        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            Stopwatch   s;
            TcpClient   c;

			for (; ; )
			{
				for (int i = 0; i < NUM_SERVERS; i++)
				{
					try
					{
						if (backgroundWorker1.CancellationPending)
							return;

						IPHostEntry dns = Dns.GetHostEntry(serverAddresses[i]);

						IPEndPoint endpoint = new IPEndPoint(dns.AddressList[0], 1935);

						s = new Stopwatch();

						ManualResetEvent ev = new ManualResetEvent(false);

						c = new TcpClient();

						s.Start();
						c.BeginConnect(dns.AddressList[0], 1935, new AsyncCallback(ConnectCallback), ev);
						
						bool success = ev.WaitOne(2500);
						
						s.Stop();
						c.Close();

						if (backgroundWorker1.CancellationPending)
							return;

						if (success)
						{
							sampler.AddSample(i, (int)s.ElapsedMilliseconds);
							backgroundWorker1.ReportProgress(2);
						}
					}
					catch (SocketException)
					{

					}

					System.Threading.Thread.Sleep(10);
				}
			}
        }

		private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			sampler.UpdateStatus(this);
		}

		private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
		{
			backgroundWorker1.CancelAsync();
		}

		public partial class myTwitchServer
		{
			public string url;
			public string name;

			public myTwitchServer (string myUrl, string myName)
			{
				url = myUrl;
				name = myName;
			}
		}

		public partial class myTwitchSorter : IComparer
		{
			int IComparer.Compare (Object x, Object y)
			{
				myTwitchServer a1 = (myTwitchServer)x;
				myTwitchServer a2 = (myTwitchServer)y;

				if (a1 == a2)
					return 0;

				// This is the default server, always sort it to the top of the list
				if (a1.url == "rtmp://live-3c.justin.tv/app")
					return -1;
				if (a2.url == "rtmp://live-3c.justin.tv/app")
					return 1;

				return( (new CaseInsensitiveComparer()).Compare( a1.name, a2.name ) );
			}
		}

		private void backgroundWorker2_DoWork(object sender, DoWorkEventArgs e)
		{
			try
			{
				XmlDocument xmlDoc = new XmlDocument();
				xmlDoc.Load(XmlReader.Create("http://api.justin.tv/api/ingest/xsplit.xml"));

				XmlNodeList locations = xmlDoc.GetElementsByTagName("locations");
				locations = locations[0].ChildNodes;

				twitchServers = new myTwitchServer[locations.Count];

				int i = 0;
				foreach (XmlNode location in locations)
				{
					string address = (string)location.Attributes["url"].Value;
					string name = (string)location.Attributes["name"].Value;

					twitchServers[i] = new myTwitchServer(address, name);
					i++;
				}

				Array.Sort(twitchServers, new myTwitchSorter());
			}
			catch (Exception)
			{
				MessageBox.Show("Failed to download or parse the Twitch TV server list. Check your internet connection.");
			}

			backgroundWorker2.ReportProgress(100);
		}

		private void backgroundWorker2_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			label1.Visible = label2.Visible = label3.Visible = label4.Visible = true;
			label5.Visible = false;

			textHosts = new TextBox[twitchServers.Length];
			labelLast = new Label[twitchServers.Length];
			labelAvg = new Label[twitchServers.Length];
			labelJitter = new Label[twitchServers.Length];
			serverAddresses = new string[twitchServers.Length];

			int vOffset = 0;
			int hOffset = 50;

			foreach (myTwitchServer server in twitchServers)
			{
				Label newLabel = new Label();
				newLabel.AutoSize = true;
				newLabel.Location = new System.Drawing.Point(157 + hOffset, 31 + vOffset);
				newLabel.Size = new System.Drawing.Size(26, 13);
				newLabel.TabIndex = 5;
				newLabel.Text = "0ms";
				this.Controls.Add(newLabel);
				labelLast[NUM_SERVERS] = newLabel;

				newLabel = new Label();
				newLabel.AutoSize = true;
				newLabel.Location = new System.Drawing.Point(208 + hOffset, 31 + vOffset);
				newLabel.Size = new System.Drawing.Size(26, 13);
				newLabel.TabIndex = 6;
				newLabel.Text = "0ms";
				this.Controls.Add(newLabel);
				labelAvg[NUM_SERVERS] = newLabel;

				newLabel = new Label();
				newLabel.AutoSize = true;
				newLabel.Location = new System.Drawing.Point(256 + hOffset, 31 + vOffset);
				newLabel.Size = new System.Drawing.Size(26, 13);
				newLabel.TabIndex = 7;
				newLabel.Text = "0ms";
				this.Controls.Add(newLabel);
				labelJitter[NUM_SERVERS] = newLabel;

				TextBox newTextBox = new TextBox();
				newTextBox.Location = new System.Drawing.Point(12, 28 + vOffset);
				newTextBox.ReadOnly = true;
				newTextBox.Size = new System.Drawing.Size(131 + 50, 20);
				newTextBox.TabIndex = 0;
				newTextBox.Text = server.name;
				this.Controls.Add(newTextBox);

				textHosts[NUM_SERVERS] = newTextBox;

				string address = server.url;
				address = address.Substring(7);
				address = address.Substring(0, address.IndexOf('/'));
				serverAddresses[NUM_SERVERS] = address;

				NUM_SERVERS++;
				vOffset += 24;
			}

			this.Height = vOffset + 64;

			sampler = new Sample(NUM_SERVERS);

			backgroundWorker1.RunWorkerAsync();

		}
    }
}
